// +---------------------------------------------------------------------------+
//
//      Program:    hott_voltage_sensor.c
//
//      Purpose:    Reads the total voltage of a 3S LiPo battery in a RC plane
//					and reports that value back to the remote control, just
//					like a commercial Hott sensor.
//      
//      Target:     ATtiny13A
//
//      Author:     Herbert Dingfelder, DL5NEG
//
//      License:    GPL 3
//
//      History:    13.12.2015  Initial Coding
//                  18.12.2015  Remote control show first reaction to sensor
//                              (still with static values taken from commercial sensor)
//					19.12.2015  All parameter tried out and works great -> the problem
//                              to get a homemade HoTT protocol sensor to work is solved
//					20.12.2015	Tested on the lab-bench, works great in all implemented feature
//								=> Version 1.0 released
//					06.01.2016	Software extension started to measure and report single cell
//								voltages for all 3 cells
//					10.01.2016	Significant code restructuring to allow for a clean implementation
//								of the single cell supervision
//								Hardware is not compatible any more! HoTT port had to be shifted to
//								free up all 3 available ADCs for measurement inputs
//					31.01.2016	Prototype hardware re-wired according to the new schematic, first
//								tests with single cell reading are successfull
//					03.02.2016	More intensive testing is also successfull
//								=> Version 2.0 released
//
// +---------------------------------------------------------------------------+

#define F_CPU 9600000
#include <util/delay.h>
#include <avr/io.h>
#include <stdio.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
#include <avr/cpufunc.h>

//#define DEVICE_SMD_1						// specify here for which hardware sample the compile is
#define DEVICE_PROTOTYPE	

// *** finetune the OSCCAL value for every ATtiny13A device! ***
// *** in function eceive_byte_via_sw_uart() uncomment the pulse output
// *** on the LED pin, then aim for 416µs for 8 fullwave pulses 
#ifdef DEVICE_PROTOTYPE
	#define OSCCAL_DEVICE_SPECIFIC	58		// this value is for the DIL8 prototype
#endif
#ifdef DEVICE_SMD_1
	#define OSCCAL_DEVICE_SPECIFIC	56		// this value is for the first SMD device
#endif


#define HOTT_SENSOR_ADR_BYTE1		0x80 	// the general sensor that is emulated has the
#define HOTT_SENSOR_ADR_BYTE2		0x8d 	//   two-byte HoTT address 0x80, 0x8d
#define HOTT_START_BYTE				0x7c 	// first byte in response telegram is always 0x7c
#define HOTT_GEN_MODULE_SIG			0x8d 	// signature is 0x8d for the general module

#define LOW_CELL_VOLTAGE_THRESHOLD	170		// low single cell voltage warning level in 20mV, we choose 3.4V (3.4V/20mV=170)

//#define DEBUG_OUT					// *** set this define to activate debug output (needs about 100 byte flash) *** 


#define HOTT_PORT PORTB				// HoTT interface is connected to PB0
#define HOTT_DDR  DDRB
#define HOTT_IN   PINB
#define HOTT_BIT  _BV(0)

#define TXD_PORT HOTT_PORT          // TXD and RXD are both connected to the one HoTT pin
#define TXD_PIN  HOTT_BIT
#define TXD_DDR  HOTT_DDR
#define RXD_PIN  HOTT_BIT
#define RXD_READ HOTT_IN

#define LIPO_TAP1_ADC_CHANNEL 1		// LiPo battery tap 1 (cell 1 to GND) is connected via divider to ADC 1
#define LIPO_TAP2_ADC_CHANNEL 2		// LiPo battery tap 2 (cell 2 to GND) is connected via divider to ADC 2
#define LIPO_TAP3_ADC_CHANNEL 3     // LiPo battery tap 3 (overall voltage) is connected via divider to ADC 3          

#define LED_PORT PORTB				// LED is connected to PB1 (for debugging only, not needed for normal operation)
#define LED_DDR  DDRB
#define LED_BIT  _BV(1)

#define DEBUG_PORT PORTB			// debug pin is PB 1 (same as LED)
#define DEBUG_DDR  DDRB
#define DEBUG_PIN  _BV(1)


// ----- macros to switch TXD to low/high, to enhance code readability -----

#define UART_TXD_LOW    (TXD_PORT &= ~(TXD_PIN) )       // macro to set TXD low
#define UART_TXD_HIGH   (TXD_PORT |=  (TXD_PIN) )       // macro to set TXD high
#define UART_OPMODE_TX	(TXD_DDR |=   (TXD_PIN) ) 		// macro to switch UART pin to output
#define UART_OPMODE_RX	(TXD_DDR &=  ~(TXD_PIN) ) 		// macro to switch UART pin to input
#define UART_RXD        (RXD_READ & (RXD_PIN) )         // macro to read RXD back
#define LED_ON          (LED_PORT |=  LED_BIT)          // macro to switch on LED
#define LED_OFF         (LED_PORT &= ~LED_BIT)          // macro to switch off LED
#define DEBUG_OUT_LOW	(DEBUG_PORT &= ~(DEBUG_PIN))	// macro to set debug output low
#define DEBUG_OUT_HIGH	(DEBUG_PORT |=  (DEBUG_PIN))	// macro to set debug output high

void reset_handler(void);
unsigned int read_adc(unsigned char adc_number);
unsigned char hex2asc(unsigned char hexbyte);
void send_byte_via_sw_uart(char byte_to_send);
int receive_byte_via_sw_uart(void);
void debug_output(uint16_t value);
void wait_for_next_uart_cycle(void);
void init_RS232_timer(void);
void read_cell_voltages(uint8_t *cell1_vlt_ptr, uint8_t *cell2_vlt_ptr, uint8_t *cell3_vlt_ptr, uint8_t *battery_vlt_ptr, uint8_t *fuel_bar_ptr);

uint16_t checksum;	// global variable checksum (helps so safe memory to have this global)

// ============================= main program starts here ======================
int main(void)
{
  
    uint8_t received_databyte;
    
    uint8_t battery_voltage=124;                        // default for battery voltage (just some value above warning threshold)
    uint8_t cell1_voltage=200;                          // default for cell voltage within a normal range (200x20mV=4.0V)
    uint8_t cell2_voltage=200;                          // default for cell voltage within a normal range (200x20mV=4.0V)
    uint8_t cell3_voltage=200;                          // default for cell voltage within a normal range (200x20mV=4.0V)
    uint8_t temperature1=20;							// default for temperature1 is 0°C (20 offset per protocol definiton)
    uint8_t temperature2=20;							// default for temperature2 is 0°C (20 offset per protocol definiton)
    uint8_t fuel_bar=50;                                // default for fuel bar graph is 50%
    uint8_t alarm_beep = 0x00;                          // default for alarm beep is off
    #ifdef DEBUG_OUT
		uint16_t debug_dummy=12345;
    #endif // #ifdef DEBUG_OUT
    
    reset_handler();                                    // do all the regular port and register settings
                                                        //   and start the needed UART counter
   
	for(;;)                                             // main loop starts here
	{

		received_databyte = receive_byte_via_sw_uart();	// wait for next character on HoTT interface
		
		if(received_databyte != HOTT_SENSOR_ADR_BYTE1 )	// if it is not the first byte of our address
			continue;									//   start a new loop immediately

		received_databyte = receive_byte_via_sw_uart();	// wait for next character on Hott interface
		
		if(received_databyte != HOTT_SENSOR_ADR_BYTE2 )	// if it is not the second byte of our address
			continue;									//   start a new loop immediately
	
		// --- now we know we have been called and shall respond with our information ---
		
		LED_ON;                                         // switch the LED on while we transmit on HoTT line
		
		_delay_ms(3);                                   // protocol needs 5ms, we have 2ms more in send_byte_via_sw_uard(), so we wait 3ms here
		
		UART_OPMODE_TX;                                 // switch HoTT pin to output mode

        checksum = 0;                                   // reset checksum at start of HoTT telegram

		send_byte_via_sw_uart(HOTT_START_BYTE);			// 1 start byte constant (always 0x7c)
		send_byte_via_sw_uart(HOTT_GEN_MODULE_SIG);		// 2 sensor id (0x8d = general module)
		send_byte_via_sw_uart(alarm_beep);				// 3 warning_beeps (triggers respective voice output), 0x10= low voltage, 0x00 = no alarm => tested ok
		send_byte_via_sw_uart(0xd0);					// 4 constant value 0xd0
		send_byte_via_sw_uart(0x00);					// 5 alarm bitmask inverted, 0 all cell voltage, 1 Battery 1, 2 Battery 2, 3 Temperature 1, 4 Temperature 2, 5 Fuel, 6 mAh, 7 Altitude
		send_byte_via_sw_uart(0x00);					// 6 alarm bitmask inverted, 0 main power current, 1 main power voltage, 2 Altitude, 3 m/s, 4 m/3s// 5 unknown// 6 unknown// 7 "ON" sign/text msg active
		send_byte_via_sw_uart(cell1_voltage);			// 7 voltage cell 1 in 20mV => tested ok
		send_byte_via_sw_uart(cell2_voltage);			// 8 voltage cell 2 in 20mV => tested ok
		send_byte_via_sw_uart(cell3_voltage);			// 9 voltage cell 3 in 20mV => tested ok
		send_byte_via_sw_uart(0x00);					// 10 voltage cell 4 in 20mV => tested ok, hard coded to 0 here since we support only 3S battery
		send_byte_via_sw_uart(0x00);					// 11 voltage cell 5 in 20mV => tested ok, hard coded to 0 here since we support only 3S battery
		send_byte_via_sw_uart(0x00);					// 12 voltage cell 6 in 20mV => tested ok, hard coded to 0 here since we support only 3S battery
		send_byte_via_sw_uart(battery_voltage);			// 13 overall voltage battery 1 in 100mV (LSB) => tested ok
		send_byte_via_sw_uart(0x00);					// 14 overall voltage battery 1 in 100mV (MSB) => tested ok (set to 0x00 here, as a 3S battery does not reach 25.6V)
		send_byte_via_sw_uart(0x00);					// 15 overall voltage battery 2 in 100mV (LSB)  => tested ok
		send_byte_via_sw_uart(0x00);					// 16 overall voltage battery 2 in 100mV (MSB)   => tested ok
		send_byte_via_sw_uart(temperature1);			// 17 temperature 1 (offset 20, 20=0°C) => tested ok
		send_byte_via_sw_uart(temperature2);			// 18 temperature 2 (offset 20, 20=0°C) => tested ok
		send_byte_via_sw_uart(fuel_bar);				// 19 fuel shown as % in bargraph (0..100) => tested ok
		#ifdef DEBUG_OUT
			send_byte_via_sw_uart( debug_dummy    &0x00ff);	// the fuel value is very convenient for debugging, at it is shown as number 
			send_byte_via_sw_uart((debug_dummy>>8)&0x00ff);	//   in the display of the LCD, we assign whatever 16bit value (up to 65000) we are interessted in to monitor
		#else
			send_byte_via_sw_uart(fuel_bar);			// 20 fuel shown in ml as text (full=65535) LSB, we use it as % number => tested ok (sidenote: value stalls at 65000) 
			send_byte_via_sw_uart(0x00);				// 21 fuel shown in ml as text (full=65535) MSB => tested ok (see line above)
		#endif // DEBUG_OUT
		send_byte_via_sw_uart(0x00);    				// 22 rpm in steps of 10 (100=1000rpm) LSB => tested ok
		send_byte_via_sw_uart(0x00);					// 23 rpm in steps of 10 (100=1000rpm) MSB => tested ok
		send_byte_via_sw_uart(0xf4);					// 24 altitude in m (500 offset, 0x01f4=500=0m) LSB => tested ok (result only shown as text)
		send_byte_via_sw_uart(0x01);					// 25 altitude in m (500 offset, 0x01fa=500=0m) MSB => tested ok (result only shown as text)
		send_byte_via_sw_uart(0x30);					// 26 climb rate in 0.01ms/s, 0x7530=30000=0.00m/s LSB => tested ok
		send_byte_via_sw_uart(0x75);					// 27 climb rate in 0.01ms/s, 0x7530=30000=0.00m/s MSB => tested ok
		send_byte_via_sw_uart(0x78);                    // 28 climb rate in m/3s, 0x78=120=0m/3s => tested ok
		send_byte_via_sw_uart(0x00);					// 29 current in 100mA, 100=10A LSB => tested ok
		send_byte_via_sw_uart(0x00);					// 30 current in 100mA, 100=10A MSB => tested ok
		send_byte_via_sw_uart(battery_voltage);			// 31 main power voltage in 100mV (shown in overview screen) LSB => tested ok
		send_byte_via_sw_uart(0x00);					// 32 main power voltage in 100mV (shown in overview screen) MSB => tested ok (set to 0x00 here, as a 3S battery does not reach 25.6V)
		send_byte_via_sw_uart(0x00);					// 33 used battery capacity in 10mAh, LSB => tested ok
		send_byte_via_sw_uart(0x00);					// 34 used battery capacity in 10mAh, MSB => tested ok
		send_byte_via_sw_uart(0x00);					// 35 speed LSB => not tested
		send_byte_via_sw_uart(0x00);					// 36 speed MSB => not tested
		send_byte_via_sw_uart(0x00);					// 37 minimum cell voltage in 20mV steps? => not tested
		send_byte_via_sw_uart(0x00);					// 38 number of the cell with the lowest voltage => not tested
		send_byte_via_sw_uart(0x00);					// 39 rpm 2 in steps of 10 (100=1000rpm) LSB => not tested
		send_byte_via_sw_uart(0x00);					// 40 rpm 2 in steps of 10 (100=1000rpm) MSB => not tested
		send_byte_via_sw_uart(0x00);					// 41 general error number (12=voice) => not tested
		send_byte_via_sw_uart(0x00);					// 42 high pressure up to 16bar in 100mbar  => not tested
		send_byte_via_sw_uart(0x00);					// 43 version number => not tested
		send_byte_via_sw_uart(0x7d);					// 44 stop byte 0x7d
        send_byte_via_sw_uart(checksum & 0xFF);			// 45 checksum (last 8bits of sum of all data bytes)
		
		_delay_ms(1);                                   // pause of 1ms before we switch back to receive mode
        UART_OPMODE_RX;                                 // switch back to receive mode
		LED_OFF;                                        // switch LED off, we are done with transmitting
		
        // --- now we can take care to get new measurement values from the ADCs and fill our status variable with the latest data ---

		read_cell_voltages(&cell1_voltage, &cell2_voltage, &cell3_voltage, &battery_voltage, &fuel_bar);	// trigger the ADC reading, returns the cell voltages in 20mV
																											// and the overall battery voltage in 100mV, and fuel bar in %

		// if one of the cell voltages gets too low...
		if(cell1_voltage <= LOW_CELL_VOLTAGE_THRESHOLD || cell2_voltage <= LOW_CELL_VOLTAGE_THRESHOLD || cell3_voltage <= LOW_CELL_VOLTAGE_THRESHOLD)
			alarm_beep = 0x10;							// trigger acoustic warning at remote control
		else
			alarm_beep = 0x00;							// no warning when voltage is ok on all cells
    
    }

    return 0;
    
}


// --------------------------------- reset handler -----------------------------
void reset_handler(void)
{
    
    CLKPR = 0x80;                       // set clock prescaler change enable 
    CLKPR = 0x00;                       // set prescaler to clock division factor 1
                                        // => system clock runs at the full 9.6MHz
    
	OSCCAL = OSCCAL_DEVICE_SPECIFIC;    // this value needs to be trimmed for every sample of ATtiny13A*
    
	init_RS232_timer();				    // start a clock counter for 19k2 baudrate
     
	LED_DDR |= LED_BIT;				    // set LED pin as output
    LED_OFF;                            // start with LED off
	
	UART_OPMODE_RX;					    // the HoTT interface pin to high impedance first (receiving)
    UART_TXD_HIGH;					    // keep the logical value on the HoTT high from the start, otherwise
                                        //   switching to TX mode will cause a falling edge right away which
                                        //   would erronously be detected as 0x00 data byte the the HoTT receiver

	#ifdef DEBUG_OUT				
		DEBUG_DDR |= DEBUG_PIN; 	    // set debug pin as output 
		DEBUG_OUT_HIGH;				    // and set it high (default for serial communication)
	#endif	// DEBUG_OUT

}

//------------------------------- read adc -----------------------
unsigned int read_adc(unsigned char adc_number)
{

    uint16_t adc_result;
    
    ADCSRA |= _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0); // ADC prescaler :128, that 
                                        // gives ADC frequency of 11.059MHz/128 = 86kHz

    ADMUX = adc_number;                 // select Vcc reference (Bits 7 and 6 remain zero)
                                        // and chose the ADC channel

    ADCSRA |= _BV(ADEN);                // switch on the ADC in general

    ADCSRA |= _BV(ADSC);                // start a single ADC conversion
    while( ADCSRA & _BV(ADSC) );        // wait until conversion is complete

    // data sheet recommends to discard first conversion, so we do one more
    ADCSRA |= _BV(ADSC);                // start a single ADC conversion
    while( ADCSRA & _BV(ADSC) );        // wait until conversion is complete

    adc_result = ADC;                   // take over ADC reading result into variable

    ADCSRA = 0;                         // completely disable the ADC to save power
    	
	return adc_result;
	
}


#ifdef DEBUG_OUT
//--------------------------- hex to asc ---------------------------------
// converts a hex value (00..0F) into the ASCII-character '0'..'F'
unsigned char hex2asc(unsigned char hexbyte)
{
    
    hexbyte += 48;                      // ASCII code for 0-9 is 48-57
    
    if(hexbyte>=58)                     // ASCII code for A-F is 65-70 -> in that case add 7
        hexbyte += 7;
    
    return hexbyte;
}
#endif //#ifdef DEBUG_OUT


// ------------------------------- send byte via uart --------------------------
void send_byte_via_sw_uart(char byte_to_send)
{
    uint8_t bit;    
    
    _delay_ms(2);						// the Hott protocol wants delays between data bytes,
										//   we place this delay generically here, rather than
										//	 for every call of this function, to save memory
    
    checksum += byte_to_send;           // to add up the checksum for the complete protocol here
                                        //   is a way to safe memory
    
    wait_for_next_uart_cycle();         // synchronize with the next overflow of the timer
    
    UART_TXD_LOW;                       // send startbit (always 0)
    
    #ifdef DEBUG_OUT
		DEBUG_OUT_LOW;
    #endif	// DEBUG_OUT
    
    for( bit=0 ; bit<=7 ; bit++ )       // send all 8 bits of this byte, starting with LSB
    {
       
        wait_for_next_uart_cycle();     // synchronize with the next overflow of the timer
        
        if(byte_to_send & _BV(0))       // take TXD line to high if LSB is set
        {
            UART_TXD_HIGH;
            #ifdef DEBUG_OUT
				DEBUG_OUT_HIGH;
			#endif	// DEBUG_OUT
        }
        else
        {
            UART_TXD_LOW; 				// respectively take it low if LSB is not set
            #ifdef DEBUG_OUT
				DEBUG_OUT_LOW;
			#endif	// DEBUG_OUT           
        }
        
        byte_to_send>>=1;               // shift the byte to be send by one bit towards LSB
    
    }   
    
    wait_for_next_uart_cycle();         // synchronize with the next overflow of the timer

    UART_TXD_HIGH;                      // send stopbit (always 1)
    
    #ifdef DEBUG_OUT
		DEBUG_OUT_HIGH;
     #endif	// DEBUG_OUT
       
}


// ------------------------------ receive byte via uart (blocking)------------------------
int receive_byte_via_sw_uart(void)
{
  
    unsigned char bit_counter;
    unsigned char byte_received=0;
  
	UART_OPMODE_RX;								// switch the joint UART pin to RX mode
    
    while( UART_RXD );                          // wait for UART RXD to go low (beginning of startbit)
												// (this is the most frequented waiting loop)
    
    TCNT0 = 0;                                  // reset counter to synchronize on incoming bits
	
                                                // read 8 data bit (plus start bit)
    for( bit_counter = 0; bit_counter < 9 ; bit_counter++ )
    {

        byte_received >>= 1;                    // shift the already received bits one right
            
        TIFR0 |= _BV(OCF0B);                    // clear output compare B match flag
        while( !(TIFR0 & _BV(OCF0B)) );         // wait for middle of bit
          
        if( UART_RXD )                          // sample on RXD, if it is high...
        {
            byte_received |= _BV(7);            // ...set MSB in received byte
        }
        
        LED_ON; _delay_us(20); LED_OFF;			// helps to finetune OSCCAL (aim for 416µs for 8 pulses)
        
    }

    // now only the stop bit is left, which we simply ignore, as it is always high
    TIFR0 |= _BV(OCF0B);                    	// clear output compare B match flag
    while( !(TIFR0 & _BV(OCF0B)) );         	// wait for middle of bit

    // intentionally we do not wait for the end of the stop bit here, that gives us
    // half a bit of time to process the received byte, before the next one may arive
    // (waiting for end of stop bit could be done with: wait_for_next_uart_cycle(); )

    return byte_received;
    
}


#ifdef DEBUG_OUT
// --- send the 16bit parameter value as readable hex number to UART for debugging ---
void debug_output(uint16_t value)
{
    
        send_byte_via_sw_uart('0');
        send_byte_via_sw_uart('x');
        send_byte_via_sw_uart(hex2asc((value>>12)&0x0f));
        send_byte_via_sw_uart(hex2asc((value>> 8)&0x0f));
        send_byte_via_sw_uart(hex2asc((value>> 4)&0x0f));
        send_byte_via_sw_uart(hex2asc((value    )&0x0f));
        send_byte_via_sw_uart(' ');    

}
#endif


// ----- read cell voltages via a 1:4.3 voltage divider each, reference is 3.34V -----
// returns cell voltages in 20mV steps, total battery voltage in 100mV and the fuel bar indicator in percent
void read_cell_voltages(uint8_t *cell1_vlt_ptr, uint8_t *cell2_vlt_ptr, uint8_t *cell3_vlt_ptr, uint8_t *battery_vlt_ptr, uint8_t *fuel_bar_ptr)
{
	
	uint16_t adc_result;
	uint16_t avg_sum1, avg_sum2, avg_sum3; 
    uint8_t a;
	uint16_t tap1_voltage, tap2_voltage, tap3_voltage;
   
    avg_sum1=0;											// reset all averaging values for a start
    avg_sum2=0;
    avg_sum3=0;
    
    for(a=0; a<8; a++)                              	// do an averaging over 8 ADC reads for each cell
    {
        adc_result = read_adc(LIPO_TAP1_ADC_CHANNEL);	// read ADC for cell1 
        avg_sum1 += adc_result;                      	// add up the results on cell1 for the averaging
        
        adc_result = read_adc(LIPO_TAP2_ADC_CHANNEL);	// read ADC for cell2 
        avg_sum2 += adc_result;                      	// add up the results on cell2 for the averaging
        
        adc_result = read_adc(LIPO_TAP3_ADC_CHANNEL);	// read ADC for cell3 
        avg_sum3 += adc_result;                      	// add up the results on cell3 for the averaging
        
        _delay_ms(1);                               	// wait 1ms for good independence between single reads
    }
    
    avg_sum1 >>= 3;                                  	// divide the added results by 8 for final result
    avg_sum2 >>= 3;                                  	// divide the added results by 8 for final result
    avg_sum3 >>= 3;                                  	// divide the added results by 8 for final result
    
    // Now we need to convert the ADC raw results into 20mV steps for each cell.
    // The division factor from ADC to 20mV units at 10bit ADC resolution, 3.34V reference and 1:4.3 divider is:
    // div_factor = 1024 * 0,020V / (3,34V * 4,3) = 1,426
    // In order to avoid float division (to safe memory), we mimic this by calculating *45/64 which gives
    // a division factor of 1,422, hence less then 0,3% error compared to the ideal calculation. That is well 
    // below the accurracy of the voltage reference and the voltage divider resistors.
    // The compiler is clever enough to replace /64 by >>6 which saves a lot of memory.
    
    tap1_voltage = (avg_sum1 * 45) / 64;				// calculate absolute voltage on tap1 in 20mV steps
    tap2_voltage = (avg_sum2 * 45) / 64;				// calculate absolute voltage on tap2 in 20mV steps
    tap3_voltage = (avg_sum3 * 45) / 64;				// calculate absolute voltage on tap3 in 20mV steps
    
    // Same procedure as described in detail above, now for the 100mV steps for the total battery voltage.
    // Since the needed unit for the overall voltage is 100mV, not 20mV, we scale up by a factor 9 instead of 45.
    
    *battery_vlt_ptr = (uint8_t)((avg_sum3 * 9) / 64);	// calculate total voltage in 100mV steps
    
    // Now we calculate the single cell voltages for the tap voltages with reference to GND
    
    *cell1_vlt_ptr = (uint8_t)tap1_voltage;						// cell voltage 1 is simply the voltage on tap 1 to GND
    *cell2_vlt_ptr = (uint8_t)tap2_voltage - *cell1_vlt_ptr;	// cell voltage 2 is voltage on tap2 minus cell voltage 1
    *cell3_vlt_ptr = (uint8_t)tap3_voltage - *cell2_vlt_ptr - *cell1_vlt_ptr;	// cell voltage 3 is voltage on tap3 minus cell voltages 1 and 2
    	
    // Finally we calculate the fuel bar percentage, where 0% corresponds to 3.2V per cell and 100% corresponds to 4.2V per cell
    // Substract 684 (ADC value at 3*3.2V), then divide by 2,14 (in the now know way as above), as ADC value at 3*4.2V ist 214 highen than at 3*3.2V)
    
    if(avg_sum3 >= 684)
        *fuel_bar_ptr = (uint8_t)(( (avg_sum3-684) * 30 ) /64 );
    else
		*fuel_bar_ptr = 0;								// everything below voltage that corresponds to 0% always has to show 0
		
    	    	
}


// ----------- synchronize with the next overflow of timer1 -----------------
void wait_for_next_uart_cycle(void)
{
    
    TIFR0 |= _BV(OCF0A);                            // clear the timer overflow flag
    while( !( TIFR0 & _BV(OCF0A)) );                // wait until overflow flag is set again

}


// --- switch on a permanent clock source that runs at the baudrate of 19200kHz
void init_RS232_timer(void)
{
    
    // baudrate is 19200 bit/s => bit length is 52.08µs
    // main clock is 9.6MHz, with prescaler 8 => clock period 833.33ns 
    // => 62.4998 clock cycles per bit length
    
    // set counter compare match to 62.5 -1 is roughly 62
    // (the -1 is due to the fact that the value is actually reached before the counter is reset)
    OCR0A = 62;        // compiler can handle the 16bit register access, simply assign the number
    
    // for sampling incoming bits in the middle, use the second compare match unit at half that value
    OCR0B = 31;
    
    // switch on the counter with prescale factor 8 
    // and switch on the clear-on-compare-match feature to reset the counter automatically when the
    // value of the compare match register A is reached
    TCCR0A = _BV(WGM01);  				// clear timer on compare with OCR0A
    TCCR0B = _BV(CS01);					// clock prescaler is 8
 
    TCNT0 = 0;                          // reset the counter, so it starts to count at zero
    
}


